<?php

/**
 * @file
 * Contains \Drupal\supercache\Cache\CacheCacheTagsChecksum.
 */

namespace Drupal\supercache\Cache;

use Drupal\Core\Cache\CacheTagsInvalidatorInterface;
use Drupal\Core\Cache\CacheTagsChecksumInterface;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Cache\CacheFactoryInterface;

/**
 * Cache tags invalidations checksum implementation that uses
 * a CacheRawBackendInterface as the storage.
 */
class CacheCacheTagsChecksum implements CacheTagsChecksumInterface, CacheTagsInvalidatorInterface {

  /**
   * Backend storage
   *
   * @var CacheRawBackendInterface
   */
  protected $backend = NULL;


  /**
   * Contains already loaded cache invalidations from the database.
   *
   * @var array
   */
  protected $tagCache = array();

  /**
   * A list of tags that have already been invalidated in this request.
   *
   * Used to prevent the invalidation of the same cache tag multiple times.
   *
   * @var array
   */
  protected $invalidatedTags = array();

  /**
   * Constructs a CacheCacheTagsChecksum object.
   */
  public function __construct(CacheRawFactoryInterface $cache) {
    $this->backend = $cache->get('supercache_tags');
  }

  /**
   * Counter implementation that works on both
   * old and new cache implementations.
   * 
   * @param string $cid
   *   The cache id.
   * @param int $increment
   *   The amount to increment the variable.
   * @param int $default
   *   Default value if the variable does not exist.
   */
  protected function counter($cid, $increment, $default = 1) {
    if ($this->backend instanceof CacheBackendInterface) {
      $data = NULL;
      if ($cache = $this->backend->get($cid)) {
        $data = $cache->data + $increment;
      }
      else {
        $data = $default;
      }
      $this->backend->set($cid, $data);
    }
    else if ($this->backend instanceof CacheRawBackendInterface) {
      $this->backend->counter($cid, $increment, $default);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function invalidateTags(array $tags) {
    foreach ($tags as $tag) {
      // Only invalidate tags once per request unless they are written again.
      if (isset($this->invalidatedTags[$tag])) {
        unset($tags[$tag]);
        continue;
      }
      $this->invalidatedTags[$tag] = TRUE;
      unset($this->tagCache[$tag]);
      $this->counter($tag, 1);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getCurrentChecksum(array $tags) {
    // Remove tags that were already invalidated during this request from the
    // static caches so that another invalidation can occur later in the same
    // request. Without that, written cache items would not be invalidated
    // correctly.
    foreach ($tags as $tag) {
      unset($this->invalidatedTags[$tag]);
    }
    return $this->calculateChecksum($tags);
  }

  /**
   * {@inheritdoc}
   */
  public function isValid($checksum, array $tags) {
    return $checksum == $this->calculateChecksum($tags);
  }

  /**
   * Calculates the current checksum for a given set of tags.
   *
   * @param array $tags
   *   The array of tags to calculate the checksum for.
   *
   * @return int
   *   The calculated checksum.
   */
  protected function calculateChecksum(array $tags) {
    $checksum = 0;

    $query_tags = array_diff($tags, array_keys($this->tagCache));
    if ($query_tags) {
      $db_tags = array();

      foreach($this->backend->getMultiple($query_tags) as $cid => $cache) {
        $db_tags[$cid] = $cache->data;
      }

      // If we were not retrieve a tag (because it has never been invalidated) make
      // sure that we initialize it, othewise the storage backend will keep looking
      // for it once again and again in the persistent backend, and that is something
      // we want to avoid!
      $missing_tags = array_diff($query_tags, array_keys($db_tags));
      foreach ($missing_tags as $tag) {
        $this->counter($tag, 1);
      }

      $this->tagCache += $db_tags;

      // Fill static cache with empty objects for tags not found in the database.
      $this->tagCache += array_fill_keys($missing_tags, 0);
    }

    foreach ($tags as $tag) {
      $checksum += $this->tagCache[$tag];
    }

    return $checksum;
  }

  /**
   * {@inheritdoc}
   */
  public function reset() {
    $this->tagCache = array();
    $this->invalidatedTags = array();
  }
}
