<?php

/**
 * @file
 * Contains \Drupal\supercache\Cache\ChainedFastBackendFactory.
 */

namespace Drupal\supercache\Cache;

use Drupal\Core\Site\Settings;
use Drupal\Core\Cache\CacheFactoryInterface;
use Drupal\Core\Cache\CacheBackendInterface;

use Symfony\Component\DependencyInjection\ContainerAwareTrait;

/**
 * Defines the chained fast cache backend factory.
 */
class ChainedFastRawBackendFactory implements CacheRawFactoryInterface {

  use ContainerAwareTrait;

  /**
   * The service name of the consistent backend factory.
   *
   * @var string
   */
  protected $consistentServiceName;

  /**
   * The service name of the fast backend factory.
   *
   * @var string
   */
  protected $fastServiceName;

  /**
   * Constructs ChainedFastBackendFactory object.
   *
   * @param Settings|NULL $settings
   *   (optional) The settings object.
   * @param string|NULL $consistent_service_name
   *   (optional) The service name of the consistent backend factory. Defaults
   *   to:
   *   - $settings->get('cache')['default'] (if specified)
   *   - 'cache.backend.database' (if the above isn't specified)
   * @param string|NULL $fast_service_name
   *   (optional) The service name of the fast backend factory. Defaults to:
   *   - 'cache.backend.apcu' (if the PHP process has APCu enabled)
   *   - NULL (if the PHP process doesn't have APCu enabled)
   */
  public function __construct(Settings $settings = NULL, $consistent_service_name = NULL, $fast_service_name = NULL) {
    // Default the consistent backend to the site's default backend.
    if (!isset($consistent_service_name)) {
      $cache_settings = isset($settings) ? $settings->get('cache') : array();
      $consistent_service_name = isset($cache_settings['default']) ? $cache_settings['default'] : 'cache.rawbackend.database';
    }

    // Default the fast backend to APCu if it's available.
    if (!isset($fast_service_name) && function_exists('apcu_fetch')) {
      $fast_service_name = 'cache.rawbackend.apcu';
    }

    $this->consistentServiceName = $consistent_service_name;
    $this->fastServiceName = $fast_service_name;
  }

  /**
   * Instantiates a chained, fast cache backend class for a given cache bin.
   *
   * @param string $bin
   *   The cache bin for which a cache backend object should be returned.
   *
   * @return CacheRawBackendInterface
   *   The cache backend object associated with the specified bin.
   */
  public function get($bin) {

    $persistentBackend = $this->container->get($this->consistentServiceName)->get($bin);

    if (empty($this->fastServiceName)) {
      return $persistentBackend;
    }

    $fastBackend = $this->container->get($this->fastServiceName)->get($bin);

    // Retrieve the number of write counts to the
    // persistent cache.
    $shared_write_count = 0;
    if ($cache = $persistentBackend->get(ChainedFastRawBackend::WRITE_COUNTER_PREFIX)) {
      $shared_write_count = $cache->data;
    }

    // Retrieve the number of write counts stored
    // in the fast cache.
    if ($cache = $fastBackend->get(ChainedFastRawBackend::WRITE_COUNTER_PREFIX)) {
      $local_write_count = $cache->data;
      if ($local_write_count != $shared_write_count) {
        // This means that we need to clear the whole binary
        // in the fast backend.
        $fastBackend->removeBin();
      }
      // This situation might happen if the persistent storage
      // is not as persistent as we think...
      if ($local_write_count > $shared_write_count) {
        $persistentBackend->set(ChainedFastRawBackend::WRITE_COUNTER_PREFIX, $local_write_count);
      }
    }

    // Use the chained backend only if there is a fast backend available;
    // otherwise, just return the consistent backend directly.
    if (isset($this->fastServiceName)) {
      return new ChainedFastRawBackend(
        $this->container->get($this->consistentServiceName)->get($bin),
        $this->container->get($this->fastServiceName)->get($bin)
      );
    }
    else {
      return $this->container->get($this->consistentServiceName)->get($bin);
    }
  }

}
